from datetime import datetime, timedelta
from peewee import fn, JOIN
from ..db import Task, Category


def get_category_summary(num: int = 5) -> list[dict]:
    """
    Returns summary of top categories with task counts by status and due dates.

    Args:
        num: Number of categories to return, ordered by total task count

    Returns:
        List of dicts containing category name and task statistics
    """
    now = datetime.now()
    week_from_now = now + timedelta(days=7)

    # Subquery for overdue tasks
    overdue_count = (
        Task.select(Task.category, fn.COUNT(Task.id).alias("overdue"))
        .where(
            (~Task.deleted) & (Task.due < now) & (Task.status != "done")
        )
        .group_by(Task.category)
    )

    # Subquery for tasks due in next 7 days
    due_soon_count = (
        Task.select(Task.category, fn.COUNT(Task.id).alias("due_soon"))
        .where(
            (~Task.deleted)
            & (Task.due >= now)
            & (Task.due <= week_from_now)
            & (Task.status != "done")
        )
        .group_by(Task.category)
    )

    # Main query joining all the information
    query = (
        Category.select(
            Category.name,
            fn.COALESCE(fn.SUM(Task.status == "zero"), 0).alias(
                "zero_count"
            ),
            fn.COALESCE(fn.SUM(Task.status == "wip"), 0).alias(
                "wip_count"
            ),
            fn.COALESCE(fn.SUM(Task.status == "blocked"), 0).alias(
                "blocked_count"
            ),
            fn.COALESCE(fn.SUM(Task.status == "done"), 0).alias(
                "done_count"
            ),
            fn.COALESCE(overdue_count.c.overdue, 0).alias("overdue"),
            fn.COALESCE(due_soon_count.c.due_soon, 0).alias("due_soon"),
        )
        .join(Task, JOIN.LEFT_OUTER)
        .join(
            overdue_count,
            JOIN.LEFT_OUTER,
            on=(Category.id == overdue_count.c.category_id),
        )
        .join(
            due_soon_count,
            JOIN.LEFT_OUTER,
            on=(Category.id == due_soon_count.c.category_id),
        )
        .where(~Task.deleted)
        .group_by(Category.id)
        .order_by(fn.COUNT(Task.id).desc())
        .limit(num)
    )

    return [
        {
            "category": cat.name,
            "tasks": {
                "zero": cat.zero_count,
                "wip": cat.wip_count,
                "blocked": cat.blocked_count,
                "done": cat.done_count,
                "overdue": cat.overdue,
                "due_soon": cat.due_soon,
            },
        }
        for cat in query
    ]


def get_recently_active(num: int = 5, category: str | None = None) -> list[dict]:
    """
    Returns most recently active tasks, optionally filtered by category.

    Args:
        num: Number of tasks to return
        category: Optional category name to filter by

    Returns:
        List of tasks ordered by last activity (updated_at)
    """
    query = (
        Task.select(Task, Category).join(Category, JOIN.LEFT_OUTER).where(~Task.deleted)
    )

    if category:
        query = query.where(Category.name == category)

    query = query.order_by(Task.updated_at.desc()).limit(num)

    return [
        {
            "id": task.id,
            "text": task.text,
            "status": task.status,
            "type": task.type,
            "category": task.category.name if task.category else None,
            "due": task.due,
            "updated_at": task.updated_at,
        }
        for task in query
    ]


def get_due_soon(
    num: int = 5, all_overdue: bool = True, category: str | None = None
) -> list[dict]:
    """
    Returns tasks ordered by due date, optionally including all overdue tasks.

    Args:
        num: Number of non-overdue tasks to return
        all_overdue: If True, returns all overdue tasks plus num more tasks
                    If False, includes overdue tasks in the count
        category: Optional category name to filter by

    Returns:
        List of tasks ordered by due date
    """
    now = datetime.now()

    # unfinished tasks w/ due date
    query = (
        Task.select(Task)
        .join(Category, JOIN.LEFT_OUTER)
        .where(
            (~Task.deleted)
            & (Task.due.is_null(False))
            & (Task.due != "")
            & (Task.status != "done")
        )
    )

    # filter by category
    if category:
        query = query.where(Category.name == category)

    if all_overdue:
        # grab all overdue & append a few more on
        overdue_tasks = list(query.where(Task.due < now).order_by(Task.due))
        upcoming_tasks = list(
            query.where(Task.due >= now).order_by(Task.due).limit(num)
        )
        tasks = overdue_tasks + upcoming_tasks
    else:
        # just most due tasks
        tasks = list(query.order_by(Task.due).limit(num))

    return [
        {
            "id": task.id,
            "text": task.text,
            "status": task.status,
            "type": task.type,
            "category": task.category.name if task.category else None,
            "due": task.due,
            "days_until_due": (task.due - now).days if task.due else None,
        }
        for task in tasks
    ]
