import json
from datetime import datetime
from peewee import fn
from peewee import Case, Value
from ..db import db, Task, Category, SavedSearch
from .. import config


def category_lookup(category):
    if category:
        category, _ = Category.get_or_create(name=category)
        return category.id


def get_task(item_id: int) -> Task:
    return Task.get_by_id(item_id)


def add_task(
    text: str,
    category: str,
    status: str,
    due: datetime | None = None,
    type: str = "",
) -> Task:
    """
    Add a new task to the database.
    Returns the created task instance.
    """
    with db.atomic():
        category_id = category_lookup(category)
        task = Task.create(
            text=text, type=type, status=status, due=due, category_id=category_id
        )
    return task


def update_task(
    item_id: int,
    **kwargs,
) -> Task:
    with db.atomic():
        if category := kwargs.pop("category", None):
            kwargs["category_id"] = category_lookup(category)
        query = Task.update(kwargs).where(Task.id == item_id)
        query.execute()
        task = Task.get_by_id(item_id)
    return task


def _parse_sort_string(sort_string, status_order):
    """
    Convert sort string like 'field1,-field2' to peewee order_by expressions.
    """
    sort_expressions = []

    if not sort_string:
        return sort_expressions

    for field in sort_string.split(","):
        is_desc = field.startswith("-")
        field_name = field[1:] if is_desc else field

        if field == "status":
            if not status_order:
                status_order = list(config.STATUSES.keys())
            # CASE statement that maps each status to its position in the order
            order_case = Case(
                Task.status,
                [(s, Value(i)) for i, s in enumerate(status_order)],
            )
            sort_expressions.append(order_case.desc() if is_desc else order_case.asc())
        elif field_name == "due_date":
            expr = fn.COALESCE(getattr(Task, field_name), datetime(3000, 12, 31))
            sort_expressions.append(expr.desc() if is_desc else expr)
        else:
            field_expr = getattr(Task, field_name)
            sort_expressions.append(field_expr.desc() if is_desc else field_expr)

    return sort_expressions


def get_tasks(
    search_text: str | None = None,
    category: int | None = None,
    statuses: tuple[str] | None = None,
    sort: str = "",
) -> list[Task]:
    query = Task.select().where(~Task.deleted)

    if search_text:
        query = query.where(fn.Lower(Task.text).contains(search_text.lower()))
    if category:
        query = query.where(Task.category == Category.get(name=category))
    if statuses:
        query = query.where(Task.status.in_(statuses))

    sort_expressions = _parse_sort_string(sort, statuses)
    query = query.order_by(*sort_expressions)

    return list(query)


def get_categories() -> list[Category]:
    return list(Category.select().order_by(Category.name))


def save_view(name: str, *, filters: dict, sort_string: str) -> SavedSearch:
    filters_json = json.dumps(filters)

    with db.atomic():
        if SavedSearch.select().where(SavedSearch.name == name).exists():
            query = SavedSearch.update(
                filters=filters_json, sort_string=sort_string
            ).where(SavedSearch.name == name)
            query.execute()
        else:
            SavedSearch.create(name=name, filters=filters_json, sort_string=sort_string)


def get_saved_view_names() -> list[str]:
    return [search.name for search in SavedSearch.select()]


def get_saved_view(name: str) -> SavedSearch:
    return SavedSearch.get(SavedSearch.name == name)
