import datetime
from textual.app import App
from textual.widgets import (
    DataTable,
    Header,
    Input,
    Static,
)
from textual.containers import Container

from ..utils import (
    remove_rich_tag,
    filter_to_string,
    get_text_from_editor,
)
from .keymodal import KeyModal
from .modals import ChoiceModal, DateModal, ConfirmModal

class NotifyValidationError(Exception):
    """will notify and continue if raised"""



class TableColumnConfig:
    def __init__(
        self,
        field: str,
        display_name: str,
        *,
        default=None,
        enable_editor=False,
        filterable=True,
        read_only=False,
    ):
        self.field = field
        self.display_name = display_name
        self.default = default
        self.enable_editor = enable_editor
        self.filterable = filterable
        self.read_only = read_only

    def preprocess(self, val):
        return val  # no-op

    def start_change(self, app, current_value):    
        if current_value.endswith(ELLIPSIS):
            app.action_start_edit()
        else:
            # default edit mode
            app._show_input("edit", current_value)


class EnumColumnConfig(TableColumnConfig):
    def __init__(
        self,
        field: str,
        display_name: str,
        enum,
        **kwargs
    ):
        super().__init__(field, display_name, **kwargs)
        self.enum = enum

    def preprocess(self, val):
        if val in self.enum:
            return val
        else:
            raise NotifyValidationError(
                f"Invalid value {val}. Use: {list(self.enum)}"
            )

    def start_change(self, app, current_value):
        # a weird hack? pass app here and correct modal gets pushed
        app.push_screen(ChoiceModal(self.enum, current_value), app.apply_change)



class DateColumnConfig(TableColumnConfig):
    def preprocess(self, val):
        try:
            return datetime.strptime(val, "%Y-%m-%d")
        except ValueError:
            raise NotifyValidationError("Invalid date format. Use YYYY-MM-DD")

    def start_change(self, app, current_value):
        app.push_screen(DateModal(current_value), app.apply_change)
ELLIPSIS = "…"



class TableEditor(App):
    CSS = """
    #footer {
        dock: bottom;
        max-height: 2;
    }
    #input_bar {
        height: 2;
        border-top: solid green;
        layout: grid;
        grid-size: 2;
        grid-columns: 10 1fr;
        display: none;
    }

    #prompt_label {
        width: 10;
        content-align: left middle;
    }
    
    #input_widget {
        width: 100%;
        margin: 0;
        border: none;
    }

    #status_bar {
      border-top: solid white;
      height: 2;
      margin: 0;
      layout: grid;
      grid-size: 2;
    }

    #left_status {
      height: 1;
      margin: 0;
      padding-left: 1;
    }
    #right_status {
      height: 1;
      margin: 0;
      padding-right: 1;
      text-align: right;
    }
    """

    BINDINGS = [
        # movement
        ("h", "cursor_left", "Left"),
        ("j", "cursor_down", "Down"),
        ("k", "cursor_up", "Up"),
        ("l", "cursor_right", "Right"),
        ("g", "cursor_top", "Top"),
        ("G", "cursor_bottom", "Bottom"),
        # filtering & editing
        ("/", "start_search", "search by name"),
        ("f", "start_filter", "filter current column"),
        ("s", "start_sort", "sort"),
        ("escape", "cancel_edit", "cancel edit"),
        # edits
        ("c", "start_change", "change current cell"),
        ("e", "start_edit", "open cell in editor"),
        ("a", "add_item", "add item"),
        ("t", "toggle_cell", "toggle enum"),
        ("d", "delete_item", "delete (must be on row mode)"),
        # other
        ("q", "quit", "quit"),
        ("?", "show_keys", "show keybindings"),
    ]

    def __init__(self):
        super().__init__()
        self.filters = {}
        self.sort_string = ""  # TODO: default sort
        self.search_query = ""
        self.saved_cursor_pos = (1, 0)
        self.save_on_move = None

    def compose(self):
        self.header = Header()
        self.table = DataTable()
        self.input_widget = Input(id="input_widget")
        self.input_label = Static("label", id="prompt_label")
        self.input_bar = Container(id="input_bar")
        self.status_bar = Container(id="status_bar")
        self.left_status = Static("LEFT", id="left_status")
        self.right_status = Static(self.sort_string, id="right_status")
        yield self.header
        yield Container(self.table)
        with Container(id="footer"):
            with self.input_bar:
                yield self.input_label
                yield self.input_widget
            with self.status_bar:
                yield self.left_status
                yield self.right_status

    def on_mount(self):
        column_names = [c.display_name for c in self.TABLE_CONFIG]
        self.table.add_columns(*column_names)
        self.refresh_data(restore_cursor=False)

    def action_cursor_left(self):
        self.table.move_cursor(column=self.table.cursor_column - 1)
        if self.table.cursor_column == 0:
            self.table.cursor_type = "row"
        self._perform_save_on_move()

    def action_cursor_right(self):
        self.table.move_cursor(column=self.table.cursor_column + 1)
        if self.table.cursor_column != 0:
            self.table.cursor_type = "cell"
        self._perform_save_on_move()

    def action_cursor_up(self):
        self.table.move_cursor(row=self.table.cursor_row - 1)
        self._perform_save_on_move()

    def action_cursor_down(self):
        self.table.move_cursor(row=self.table.cursor_row + 1)
        self._perform_save_on_move()

    def action_cursor_top(self):
        self.table.move_cursor(row=0)
        self._perform_save_on_move()

    def action_cursor_bottom(self):
        self.table.move_cursor(row=self.table.row_count - 1)
        self._perform_save_on_move()

    def refresh_data(self, *, restore_cursor=True):
        # reset table
        self.table.clear()
        self.refresh_items()

        # update footer
        filter_display = filter_to_string(self.filters, self.search_query)
        self.left_status.update(f"{self.table.row_count} |{filter_display}")

        # restore cursor
        if restore_cursor:
            self.table.move_cursor(
                row=self.saved_cursor_pos[0], column=self.saved_cursor_pos[1]
            )
        else:
            self.table.move_cursor(row=0, column=1)

    def action_delete_item(self):
        self.push_screen(ConfirmModal(f"delete ?"), self._delete_item_callback)

    def _delete_item_callback(self, confirm):
        if confirm and self.table.cursor_column == 0:
            cur_row = self.table.cursor_row
            item_id = int(self.table.get_cell_at((cur_row, 0)))
            # deletable items need a delete
            self.update_item_callback(item_id, deleted=True)
            self._save_cursor()
            self.refresh_data()

    def action_add_item(self):
        # the new item should use the selected value if filtered
        # prepopulate with either the field default or the current filter
        prepopulated = {}
        for fc in self.TABLE_CONFIG:
            if fc.read_only:
                continue
            val = self.filters.get(fc.field, fc.default)
            if val is not None:
                if "," in val:
                    val = val.split(",")[0] # TODO: fix hack for enums
                prepopulated[fc.field] = val

        new_item = self.add_item_callback(**prepopulated)
        self.refresh_data(restore_cursor=False)
        self.move_cursor_to_item(new_item.id)  # TODO: check success here
        self.action_start_change()

    def _active_column_config(self):
        cur_col = self.table.cursor_column
        return self.TABLE_CONFIG[cur_col]

    def _active_item_id(self):
        return int(self.table.get_cell_at((self.table.cursor_row, 0)))

    # def action_toggle_cell(self):
    #     cur_row = self.table.cursor_row
    #     cur_col = self.table.cursor_column
    #     cconf = self._active_column_config()

    #     if cconf.enum:
    #         item_id = self._active_item_id()
    #         current_val = self.table.get_cell_at((cur_row, cur_col))
    #         next_val = advance_enum_val(cconf.enum, current_val)
    #         self.table.update_cell_at((cur_row, cur_col), next_val)
    #         # trigger item_id to be saved on the next cursor move
    #         # this avoids filtered columns disappearing right away
    #         # and tons of DB writes
    #         update = {cconf.field: next_val}
    #         self._register_save_on_move(item_id, **update)

    def _register_save_on_move(self, item_id, **kwargs):
        if self.save_on_move and self.save_on_move["item_id"] != item_id:
            # we should only ever overwrite the same item
            raise Exception("invalid save_on_move state")
        self.save_on_move = {"item_id": item_id, **kwargs}

    def _perform_save_on_move(self):
        if self.save_on_move:
            self.update_item_callback(**self.save_on_move)
            self._save_cursor()
            self.refresh_data()
            self.save_on_move = None

    def move_cursor_to_item(self, item_id):
        # ick, but only way to search table?
        for row in range(self.table.row_count):
            data = self.table.get_row_at(row)
            if data[0] == str(item_id):
                self.table.move_cursor(row=row, column=1)
                break

    # Control of edit bar ####################

    def _show_input(self, mode, start_value):
        self.mode = mode
        self.input_bar.display = True
        self.input_widget.value = start_value
        # TODO: a better solution here, violating Liskov
        if mode.endswith("-view"):
            mode_input_label = "view name: "
        else:
            mode_input_label = f"{mode}: "
        self.input_label.update(mode_input_label)
        self.set_focus(self.input_widget)

    def _hide_input(self):
        self.input_bar.display = False
        self.set_focus(self.table)

    def action_cancel_edit(self):
        self._hide_input()
        self.table.cursor_type = "cell"

    def action_start_search(self):
        self._show_input("search", "")

    def action_start_filter(self):
        cconf = self._active_column_config()
        if not cconf.filterable:
            return
        cur_filter_val = self.filters.get(cconf.field) or ""
        self._show_input("filter", cur_filter_val)
        self.table.cursor_type = "column"

    def action_start_sort(self):
        self._show_input("sort", self.sort_string)

    def _save_cursor(self):
        self.saved_cursor_pos = (self.table.cursor_row, self.table.cursor_column)

    def action_start_change(self):
        # invalid position
        if self.table.cursor_row is None or self.table.cursor_column == 0:
            return

        # check if editable
        cconf = self._active_column_config()
        if cconf.read_only:
            return

        # save cursor before callback, so correct position updates
        self._save_cursor()
        current_value = self.table.get_cell_at(
            (self.table.cursor_row, self.table.cursor_column)
        )
        current_value = remove_rich_tag(current_value)

        # delegate to start_change, which will call back to self
        cconf.start_change(self, current_value)

    def action_start_edit(self):
        cconf = self._active_column_config()

        if not cconf.enable_editor:
            return

        self._save_cursor()
        # get item from db to ensure we have the original text
        # since the table might have an edited/truncated version
        item_id = self._active_item_id()
        item = self.get_item_callback(item_id)

        # found at https://github.com/Textualize/textual/discussions/1654
        self._driver.stop_application_mode()
        new_text = get_text_from_editor(item.text)
        self.refresh()
        self._driver.start_application_mode()
        if new_text is not None:
            self.update_item_callback(item_id, text=new_text)
            self.refresh_data()

    def on_input_submitted(self, event: Input.Submitted):
        self._hide_input()
        if self.mode == "search":
            self.search_query = event.value
            self.refresh_data(restore_cursor=False)
        elif self.mode == "filter":
            cconf = self._active_column_config()
            self.filters[cconf.field] = event.value
            self.refresh_data(restore_cursor=False)
            self.table.cursor_type = "cell"
        elif self.mode == "sort":
            self.sort_string = event.value
            self.refresh_data(restore_cursor=False)
            self.right_status.update(self.sort_string)
        elif self.mode == "edit":
            self.apply_change(event.value)
        else:
            raise ValueError(f"unknown mode: {self.mode}")

    def apply_change(self, new_value):
        row, col = self.saved_cursor_pos
        item_id = int(self.table.get_cell_at((row, 0)))
        cconf = self.TABLE_CONFIG[col]
        field = cconf.field
        update_data = {}

        # preprocess/validate the field being saved
        try:
            update_data[field] = cconf.preprocess(new_value)
            self.update_item_callback(item_id, **update_data)
            self.refresh_data()
        except NotifyValidationError as e:
            self.notify(str(e))
        except Exception as e:
            self.notify(f"Error updating item: {str(e)}")
        finally:
            # break out of edit mode
            self.action_cancel_edit()

    def action_show_keys(self):
        self.push_screen(KeyModal())

    def refresh_items(self):
        """
        for now, override this method with something that
        calls self.table.add_row()

        a better interface would be to have controller
        handle the conversion from db -> display
        """
        raise NotImplementedError
