from textual.screen import ModalScreen
from textual.binding import Binding
from textual.widgets import RadioSet, RadioButton, Label
from .. import config
from ..utils import get_color_enum

class ConfirmModal(ModalScreen):
    CSS = """
    ConfirmModal {
        align: center middle;
        background: $primary 30%;
    }
    """

    BINDINGS = [
        ("y", "confirm", "Down"),
        ("n,escape", "cancel", "cancel"),
    ]

    def __init__(self, message):
        self.message = message
        super().__init__()

    def compose(self):
        yield Label(self.message)
        yield Label("(y)es")
        yield Label("(n)o")

    def action_confirm(self):
        self.dismiss(True)

    def action_cancel(self):
        self.app.pop_screen()


class ChoiceModal(ModalScreen):
    CSS = """
    ChoiceModal {
        align: center middle;
        background: $primary 30%;
    }
    ChoiceModal Label {
        height: 1;
    }
    """

    BINDINGS = [
        ("j", "cursor_down", "Down"),
        ("k", "cursor_up", "Up"),
        Binding("enter", "select", "Select", priority=True),
        ("c", "select", "Select"),
        ("escape", "cancel", "cancel"),
    ]

    def __init__(self, enum, selected):
        self._enum = enum
        self.selected = selected
        super().__init__()

    def compose(self):
        yield RadioSet(
            *[
                RadioButton(
                    get_color_enum(e.value, config.STATUSES, "red"), value=self.selected == str(e.value)
                )
                for e in self._enum
            ]
        )

    def action_cursor_down(self):
        self.query_one(RadioSet).action_next_button()

    def action_cursor_up(self):
        self.query_one(RadioSet).action_previous_button()

    def action_select(self):
        rs = self.query_one(RadioSet)
        rs.action_toggle_button()
        pressed = rs.pressed_button
        self.dismiss(str(pressed.label))

    def action_cancel(self):
        self.app.pop_screen()


class DateModal(ModalScreen):
    CSS = """
    DateModal {
        layout: horizontal;
        align: center middle;
        background: $primary 30%;
    }
    DateModal Label {
        border: solid grey;
    }
    DateModal Label.selected-date {
        border: solid green;
    }
    """

    BINDINGS = [
        ("j", "cursor_down", "Down"),
        ("k", "cursor_up", "Up"),
        ("h", "cursor_left", "Left"),
        ("l", "cursor_right", "Right"),
        #        ("0,1,2,3,4,5,6,7,8,9", "num_entry", "#"),
        Binding("enter", "select", "Select", priority=True),
        ("escape", "cancel", "cancel"),
    ]

    def __init__(self, date):
        self.pieces = [int(p) for p in date.split("-")]
        self.selected = 1  # start on month
        super().__init__()

    def compose(self):
        for idx, piece in enumerate(self.pieces):
            yield Label(
                str(piece), classes="selected-date" if idx == self.selected else ""
            )

    def action_cursor_left(self):
        # cycle Y/M/D
        self.selected = (self.selected - 1) % 3
        self._update_highlight()

    def action_cursor_right(self):
        self.selected = (self.selected + 1) % 3
        self._update_highlight()

    def _update_highlight(self):
        for idx, lbl in enumerate(self.query("Label")):
            if idx == self.selected:
                lbl.add_class("selected-date")
            else:
                lbl.remove_class("selected-date")

    def max_for(self, piece):
        days_in_month = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        if piece == 0:
            return 3000
        elif piece == 1:
            return 12
        else:
            # -1 for offset array
            return days_in_month[self.pieces[1] - 1]

    def _move_piece(self, by):
        cur_value = self.pieces[self.selected]
        cur_value += by
        if cur_value == 0:
            cur_value = self.max_for(self.selected)
        if cur_value > self.max_for(self.selected):
            cur_value = 1
        self.pieces[self.selected] = cur_value
        cur_label = self.query("Label")[self.selected]
        cur_label.update(str(cur_value))

    def action_cursor_down(self):
        self._move_piece(-1)

    def action_cursor_up(self):
        self._move_piece(1)

    def on_key(self, event) -> None:
        key = event.key
        if key in "0123456789":
            cur_value = self.pieces[self.selected]
            # If we can append the values like 1+2 = 12
            # then do that, otherwise reset the value
            # so (for months) 1+3 = 3
            appended = int(str(cur_value) + key)
            if appended <= self.max_for(self.selected):
                cur_value = appended
            else:
                cur_value = int(key)
            self.pieces[self.selected] = cur_value
            self._move_piece(0)
            event.prevent_default()

    def action_select(self):
        date = "-".join(str(p) for p in self.pieces)
        self.dismiss(date)

    def action_cancel(self):
        self.app.pop_screen()
