import json
from textual.widgets import Input
from datetime import datetime

from .. import config
from ..controller.tasks import (
    get_task,
    get_tasks,
    add_task,
    update_task,
    save_view,
    get_saved_view,
)
from ..utils import (
    get_color_enum,
    get_colored_date,
)
from .editor import (
    TableEditor,
    TableColumnConfig,
    EnumColumnConfig,
    DateColumnConfig,
    ELLIPSIS,
)


class TT(TableEditor):
    TABLE_CONFIG = (
        TableColumnConfig("id", "ID"),
        TableColumnConfig("text", "Task", default="new task", enable_editor=True),
        EnumColumnConfig(
            "status",
            "Status",
            enum=config.STATUSES,
            default="zero",
        ),
        TableColumnConfig("type", "Type", default=""),
        DateColumnConfig("due", "Due", default=""),
        TableColumnConfig("category", "Category", default="main"),
    )

    BINDINGS = [
        # saved views
        ("ctrl+s", "save_view", "save current view"),
        ("ctrl+o", "load_view", "load saved view"),
    ]

    def __init__(self, default_view="default"):
        super().__init__()
        self._load_view(default_view)
        self.update_item_callback = update_task
        self.add_item_callback = add_task
        self.get_item_callback = get_task

    def _load_view(self, name):
        try:
            saved = get_saved_view(name)
            self.filters = json.loads(saved.filters)
            self.sort_string = saved.sort_string
        except Exception:
            self.notify(f"Could not load {name}")

    def action_save_view(self):
        self._show_input("save-view", "default")

    def action_load_view(self):
        self._show_input("load-view", "")

    def on_input_submitted(self, event: Input.Submitted):
        # Override to add save/load view
        if self.mode == "save-view":
            save_view(event.value, filters=self.filters, sort_string=self.sort_string)
            self._hide_input()
            event.prevent_default()
        elif self.mode == "load-view":
            self._load_view(event.value)
            self.refresh_tasks(restore_cursor=False)
            self._hide_input()
            event.prevent_default()
        # if event isn't handled here it will bubble to parent

    def refresh_items(self):
        items = get_tasks(
            self.search_query,
            category=self.filters.get("category"),
            statuses=self.filters.get("status", "").split(",")
            if "status" in self.filters
            else None,
            sort=self.sort_string,
        )
        for item in items:
            category = get_color_enum(
                item.category.name if item.category else " - ",
                config.PROJECTS,
                "grey"
            )
            status = get_color_enum(item.status, config.STATUSES, "red")
            due = get_colored_date(item.due)

            if "\n" in item.text:
                text = item.text.split("\n")[0] + ELLIPSIS
            else:
                text = item.text

            self.table.add_row(
                str(item.id),
                text,
                status,
                item.type,
                due,
                category,
                key=str(item.id),
            )


def run(default_view):
    app = TT(default_view)
    app.run()


if __name__ == "__main__":
    run()
